#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Deep neural model
Ubuntu 16.04 & PyTorch 1.2
Last update: KzXuan, 2019.08.15
"""
__name__ = "dnnnlp"
__version__ = "1.0.3"


# verbose for output portion
class verbose():
    @classmethod
    def config(cls, level):
        """Set verbose level for the whole package.

        Args:
            level [int]: verbose level, needs 0/1/2, higher level means more output
        """
        cls.level = level

    @staticmethod
    def check(level):
        """Check a set level of one output.

        Args:
            level [int]: verbose level of a function, needs 0/1/2

        Returns:
            True or False
        """
        if verbose.level >= level:
            return True
        else:
            return False


# set default verbose level
verbose.config(2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Some common models for deep neural network.
Last update: KzXuan, 2019.08.21
"""
import torch
import numpy as np
import torch.nn as nn
import torch.utils.data as Data
import torch.nn.functional as F
from . import layer


class CNNModel(nn.Module):
    def __init__(self, args, emb_matrix=None, kernel_widths=[2, 3, 4], act_fun=nn.ReLU):
        """Initilize CNN model data and layer.

        Args:
            args [dict]: all model arguments
            emb_matrix [np.array]: word embedding matrix
            kernel_widths [list]: list of kernel widths for CNN kernel
            act_fun [torch.nn.modules.activation]: activation function
        """
        super(CNNModel, self).__init__()

        self.emb_mat = layer.EmbeddingLayer(emb_matrix, args.emb_type)
        self.drop_out = nn.Dropout(args.drop_prob)

        self.cnn = nn.ModuleList()
        for kw in kernel_widths:
            self.cnn.append(
                layer.CNNLayer(args.emb_dim, 1, args.n_hidden, kw, act_fun, args.drop_prob)
            )

        self.predict = layer.SoftmaxLayer(args.n_hidden * len(kernel_widths), args.n_class)

    def forward(self, inputs, mask=None):
        """Forward propagation.

        Args:
            inputs [tensor]: input tensor (batch_size * max_seq_len * input_size)
            mask [tensor]: mask matrix (batch_size * max_seq_len)

        Returns:
            pred [tensor]: predict of the model (batch_size * n_class)
        """
        inputs = self.emb_mat(inputs)
        assert inputs.dim() == 3, "Dimension error of 'inputs', check args.emb_type & emb_dim."
        if mask is not None:
            assert inputs.shape[:2] == mask.shape, "Dimension match error of 'inputs' and 'mask'."

        outputs = self.drop_out(inputs)
        outputs = torch.cat([c(outputs, mask, out_type='max') for c in self.cnn], -1)

        pred = self.predict(outputs)
        return pred


class RNNModel(nn.Module):
    def __init__(self, args, emb_matrix=None, n_hierarchy=1, n_layer=1,
                 bi_direction=True, rnn_type='LSTM', use_attention=False):
        """Initilize RNN model data and layer.

        Args:
            args [dict]: all model arguments
            emb_matrix [np.array]: word embedding matrix
            n_hierarchy [int]: number of model hierarchy
            n_layer [int]: number of RNN layer in a hierarchy
            bi_direction [bool]: use bi-directional model or not
            rnn_type [str]: choose rnn type with 'tanh'/'LSTM'/'GRU'
            use_attention [bool]: use attention layer
        """
        super(RNNModel, self).__init__()

        self.n_hierarchy = n_hierarchy
        self.bi_direction_num = 2 if bi_direction else 1
        self.use_attention = use_attention

        self.emb_mat = layer.EmbeddingLayer(emb_matrix, args.emb_type)
        self.drop_out = nn.Dropout(args.drop_prob)

        rnn_params = (args.n_hidden, n_layer, bi_direction, rnn_type, args.drop_prob)
        self.rnn = nn.ModuleList([layer.RNNLayer(args.emb_dim, *rnn_params)])
        if use_attention:
            self.att = nn.ModuleList(
                [layer.SoftAttentionLayer(self.bi_direction_num * args.n_hidden)]
            )
        for _ in range(self.n_hierarchy - 1):
            self.rnn.append(layer.RNNLayer(self.bi_direction_num * args.n_hidden, *rnn_params))
            if use_attention:
                self.att.append(layer.SoftAttentionLayer(self.bi_direction_num * args.n_hidden))

        self.predict = layer.SoftmaxLayer(self.bi_direction_num * args.n_hidden, args.n_class)

    def forward(self, inputs, mask=None):
        """Forward propagation.

        Args:
            inputs [tensor]: input tensor (batch_size * max_seq_len * input_size)
            mask [tensor]: mask matrix (batch_size * max_seq_len)

        Returns:
            pred [tensor]: predict of the model (batch_size * n_class)
        """
        inputs = self.emb_mat(inputs)
        if mask is not None:
            assert inputs.shape[:-1] == mask.shape, "Dimension match error of 'inputs' and 'mask'."

        _, *max_seq_len, _ = inputs.size()
        max_seq_len = max_seq_len[::-1]
        assert len(max_seq_len) == self.n_hierarchy, "Hierarchy match error of 'inputs'."

        outputs = self.drop_out(inputs)
        for hi in range(self.n_hierarchy):
            outputs = outputs.reshape(-1, max_seq_len[hi], outputs.size(-1))
            mask = mask.reshape(-1, max_seq_len[hi])

            if self.use_attention:
                outputs = self.rnn[hi](outputs, mask, out_type='all')
                outputs = self.att[hi](outputs, mask)
            else:
                outputs = self.rnn[hi](outputs, mask, out_type='last')

            mask = (mask.sum(-1) != 0).int()

        pred = self.predict(outputs)
        return pred


class TransformerModel(nn.Module):
    def __init__(self, args, emb_matrix=None, n_layer=6, n_head=8):
        """Initilize transfomer model data and layer.

        Args:
            args [dict]: all model arguments
            emb_matrix [np.array]: word embedding matrix
            n_layer [int]: number of RNN layer in a hierarchy
        """
        super(TransformerModel, self).__init__()

        self.n_layer = n_layer

        self.emb_mat = layer.EmbeddingLayer(emb_matrix, args.emb_type)
        self.drop_out = nn.Dropout(args.drop_prob)

        self.trans = nn.ModuleList(
            [layer.TransformerLayer(args.emb_dim, n_head) for _  in range(n_layer)]
        )
        self.predict = layer.SoftmaxLayer(args.emb_dim, args.n_class)

    def forward(self, inputs, mask=None):
        """Forward propagation.
        Args:
            inputs [tensor]: input tensor (batch_size * max_seq_len * input_size)
            mask [tensor]: mask matrix (batch_size * max_seq_len)

        Returns:
            pred [tensor]: predict of the model (batch_size * n_class)
        """
        inputs = self.emb_mat(inputs)
        assert inputs.dim() == 3, "Dimension error of 'inputs', check args.emb_type & emb_dim."
        if mask is not None:
            assert inputs.shape[:2] == mask.shape, "Dimension match error of 'inputs' and 'mask'."

        outputs = self.drop_out(inputs)
        for li in range(self.n_layer - 1):
            outputs = self.trans[li](outputs, query_mask=mask, out_type='all')
        outputs = self.trans[-1](outputs, query_mask=mask, out_type='first')

        pred = self.predict(outputs)
        return pred



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
classifyution functions for deep neural models.
Last update: KzXuan, 2019.08.21
"""
import time
import torch
import dnnnlp
import argparse
import numpy as np
import torch.nn as nn
import torch.utils.data as Data
import torch.multiprocessing as mp
from copy import deepcopy
from . import utils, layer, model


def set_seed(seed=100):
    """Set random seed.

    Args:
        seed [int]: seed number
    """
    torch.manual_seed(100)
    torch.cuda.manual_seed(100)
    torch.cuda.manual_seed_all(100)


def default_args():
    """Set default arguments.

    Returns:
        args [argparse.Namespace]: default arguments which can be pri-set in shell
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--n_gpu", default=1, type=int, help="number of GPUs for running")
    parser.add_argument("--space_turbo", default=True, type=bool, help="use more space to fasten")
    parser.add_argument("--data_shuffle", default=True, type=bool, help="shuffle data")
    parser.add_argument("--emb_type", default=None, type=str, help="embedding type")
    parser.add_argument("--emb_dim", default=300, type=int, help="embedding dimension")
    parser.add_argument("-c", "--n_class", default=2, type=int, help="classify classes number")

    parser.add_argument("--n_hidden", default=50, type=int, help="hidden layer nodes")
    parser.add_argument("-lr", "--learning_rate", default=0.01, type=float, help="learning rate")
    parser.add_argument("-l2", "--l2_reg", default=1e-6, type=float, help="l2 regularization")
    parser.add_argument("-b", "--batch_size", default=128, type=int, help="batch size")
    parser.add_argument("-i", "--iter_times", default=30, type=int, help="iteration times")
    parser.add_argument("--display_step", default=2, type=int, help="display inteval")
    parser.add_argument("--drop_prob", default=0.1, type=float, help="drop out ratio")
    parser.add_argument("--eval_metric", default='accuracy', type=str, help="evaluation metric")

    args = parser.parse_args()
    return args


class exec(object):
    def __init__(self, args):
        """Initilize the model argument.

        Args:
            args [argparse.Namespace]: set arguments including:
                n_gpu [int]: number of GPUs for running
                space_turbo [bool]: use more space to fasten
                data_shuffle [bool]: shuffle data
                emb_type [str]: use None/'const'/'variable'/'random'
                emb_dim [int]: embedding dimension
                n_class [int]: number of object classify classes
                n_hidden [int]: number of hidden layer nodes
                learning_rate [float]: learning rate
                l2_reg [float]: L2 regularization parameter
                batch_size [int]: train batch size
                iter_times [int]: iteration times
                display_step [int]: the interval iterations to display
                drop_prob [float]: drop out ratio
                eval_metric [str]: use 'macro'/'class1'/'acc'/...
        """
        self.args = args
        self.n_gpu = args.n_gpu
        self.space_turbo = args.space_turbo
        self.data_shuffle = args.data_shuffle
        self.emb_type = args.emb_type
        self.emb_dim = args.emb_dim
        self.n_class = args.n_class
        self.n_hidden = args.n_hidden
        self.learning_rate = args.learning_rate
        self.l2_reg = args.l2_reg
        self.batch_size = args.batch_size
        self.iter_times = args.iter_times
        self.display_step = args.display_step
        self.drop_prob = args.drop_prob
        self.eval_metric = args.eval_metric
        if self.eval_metric == 'micro':
            self.eval_metric = 'accuracy'

    def attributes_from_dict(self, args):
        """Set attributes' name and value from dict.

        Args:
            args [dict]: dict including name and value of parameters
        """
        for name, value in args.items():
            setattr(self, name, value)

    def create_data_loader(self, *data):
        """Create data loader for pytorch.

        Args:
            data [tensor]: several tensors with the same shape[0]

        Returns:
            loader [DataLoader]: torch data generator
        """
        dataset = Data.TensorDataset(*data)
        loader = Data.DataLoader(
            dataset=dataset,
            shuffle=self.data_shuffle,
            batch_size=self.batch_size
        )
        return loader


class Classify(exec):
    def __init__(self, model, args, train_x, train_y, train_mask,
                 test_x=None, test_y=None, test_mask=None):
        """Initilize classification method.

        Args:
            model [nn.Module]: a standart pytorch model
            args [dict]: all model arguments
            train_x [np.array/tensor]: training data
            train_y [np.array/tensor]: training label
            train_mask [np.array/tensor]: training mask
            test_x [np.array/tensor]: testing data
            test_y [np.array/tensor]: testing label
            test_mask [np.array/tensor]: testing mask
        """
        exec.__init__(self, args)

        self.model = model

        self.train_x = torch.as_tensor(train_x, dtype=torch.float)
        self.train_y = torch.as_tensor(train_y, dtype=torch.long)
        self.train_mask = torch.as_tensor(train_mask, dtype=torch.int)
        self.test_x = torch.as_tensor(test_x, dtype=torch.float) if test_x else test_x
        self.test_y = torch.as_tensor(test_y, dtype=torch.long) if test_y else test_y
        self.test_mask = torch.as_tensor(test_mask, dtype=torch.int) if test_mask else test_mask

    def _model_initilize(self, device_id):
        """Given initilize fuction for model.

        Args:
            device_id [int]: CPU device for -1, and GPU device for 0/1/...
        """
        assert torch.cuda.device_count() >= self.n_gpu, "Not enough GPU devices."
        self.device_id = device_id
        if self.device_id == -1:
            self.n_gpu = 0
            self.device = torch.device('cpu')
        elif self.n_gpu and self.space_turbo:
            self.device = torch.device(self.device_id)
        else:
            self.device = torch.device('cpu')

        # set cuda and parallel
        assert torch.cuda.device_count() >= self.device_id + self.n_gpu, "Not enough GPU devices."
        if self.n_gpu:
            self.model.to(self.device_id)
            if self.n_gpu > 1:
                self.gpu_dist = range(self.device_id, self.device_id + self.n_gpu)
                self.model = nn.DataParallel(self.model, device_ids=self.gpu_dist)

        # get model initial parameters
        self.model_init = deepcopy(self.model.state_dict())
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate, weight_decay=self.l2_reg
        )
        self.loss_function = nn.NLLLoss()

        # move data to initial device
        if self.n_gpu and self.space_turbo:
            self.train_x = self.train_x.to(self.device_id)
            self.train_y = self.train_y.to(self.device_id)
            self.train_mask = self.train_mask.to(self.device_id)
            if self.test_x:
                self.test_x = self.test_x.to(self.device_id)
                self.test_y = self.test_y.to(self.device_id)
                self.test_mask = self.test_mask.to(self.device_id)

    def _run_train(self, train_loader, **model_params):
        """Training part.

        Args:
            train_loader [DataLoader]: train data generator
            model_params [parameter]: more parameters for model

        Returns:
            losses [float]: loss of one iteration
        """
        self.model.train()

        loss = 0.0
        for step, (x, y, m) in enumerate(train_loader):
            if not self.space_turbo and self.n_gpu:
                x, y, m = x.to(self.device_id), y.to(self.device_id), m.to(self.device_id)
            pred = self.model(x, m, **model_params)
            batch_loss = self.loss_function(pred, y)

            self.optimizer.zero_grad()
            batch_loss.backward()
            self.optimizer.step()

            loss += batch_loss.cpu().data.numpy()
        loss = loss / (step + 1)
        return loss

    def _run_test(self, test_loader, **model_params):
        """Testing part.

        Args:
            test_loader [DataLoader]: test data generator
            model_params [parameter]: more parameters for model

        Returns:
            preds [np.array]: predicts of the test data
            tys [np.array]: true label of the test data
        """
        self.model.eval()

        preds, tys = torch.FloatTensor(), torch.LongTensor()
        for _, (tx, ty, tm) in enumerate(test_loader):
            if not self.space_turbo and self.n_gpu:
                tx, ty, tm = tx.to(self.device_id), ty.to(self.device_id), tm.to(self.device_id)
            pred = self.model(tx, tm, **model_params)

            preds = torch.cat((preds, pred.cpu().data))
            tys = torch.cat((tys, ty.cpu().data))

        preds = np.argmax(preds.data.numpy(), axis=-1)
        return preds, tys.data.numpy()

    def train_test(self, device_id=0):
        """Run training and testing.

        Args:
            device_id [int]: CPU device for -1, and GPU device for 0/1/...

        Returns:
            best_evals [dict]: the best evaluation metrics
        """
        assert self.test_x and self.test_y and self.test_mask, ValueError("Test x or y or mask may not exist.")
        self._model_initilize(device_id)

        train_loader = self.create_data_loader(
            self.train_x, self.train_y, self.train_mask
        )
        test_loader = self.create_data_loader(
            self.test_x, self.test_y, self.test_mask
        )
        self.model.load_state_dict(self.model_init)

        results = []
        ptable = utils.display_prfacc(self.eval_metric, verbose=2)
        for it in range(1, self.iter_times + 1):
            loss = self._run_train(train_loader)
            pred, ty = self._run_test(test_loader)

            evals = utils.prfacc(ty, pred, one_hot=False)
            evals.update({'iter': it, 'loss': loss})
            if it % self.display_step == 0:
                ptable.row(evals)
            results.append(evals)

        best_evals = utils.maximum_prfacc(*results, eval_metric=self.eval_metric)
        ptable = utils.display_prfacc(self.eval_metric, verbose=1)
        ptable.row(dict(best_evals, **{"iter": 'BEST'}))
        return best_evals

    def train_itself(self, device_id=0):
        """Run testing with training data.

        Args:
            device_id [int]: CPU device for -1, and GPU device for 0/1/...

        Returns:
            best_evals [dict]: the best evaluation metrics
        """
        self.test_x = self.train_x
        self.test_y = self.train_y
        self.test_mask = self.train_mask
        return self.train_test(device_id)

    def cross_validation(self, fold=10, device_id=0):
        """Run cross validation.

        Args:
            fold [int]: k fold control
            device_id [int]: CPU device for -1, and GPU device for 0/1/...

        Returns:
            best_evals [dict]: the best evaluation metrics
        """
        self._model_initilize(device_id)
        kf_avg = []

        for count, train, test in utils.mod_fold(self.train_x.shape[0], fold=fold):
            train_loader = self.create_data_loader(
                self.train_x[train], self.train_y[train], self.train_mask[train]
            )
            test_loader = self.create_data_loader(
                self.train_x[test], self.train_y[test], self.train_mask[test]
            )

            if dnnnlp.verbose.check(2):
                print()
                print("Fold {}, label {}".format(count, self.train_y[test].bincount().cpu().numpy()))
            self.model.load_state_dict(self.model_init)

            results = []
            ptable = utils.display_prfacc(self.eval_metric, verbose=2)
            for it in range(1, self.iter_times + 1):
                loss = self._run_train(train_loader)
                pred, ty = self._run_test(test_loader)

                evals = utils.prfacc(ty, pred, one_hot=False)
                evals.update({'iter': it, 'loss': loss})
                if it % self.display_step == 0:
                    ptable.row(evals)
                results.append(evals)

            best_evals = utils.maximum_prfacc(*results, eval_metric=self.eval_metric)
            ptable.line()
            ptable.row(dict(best_evals, **{"iter": 'BEST'}))

            kf_avg.append(best_evals)

        avg_evals = utils.average_prfacc(*kf_avg)
        if dnnnlp.verbose.check(1):
            print()
        ptable = utils.display_prfacc(self.eval_metric, verbose=1)
        ptable.row(dict(avg_evals, **{"iter": 'AVG'}))
        return avg_evals


class SequenceLabeling(Classify):
    def __init__(self, model, args, train_x, train_y, train_mask,
                 test_x=None, test_y=None, test_mask=None):
        Classify.__init__(
            model, args, train_x, train_y, train_mask, test_x, test_y, test_mask
        )


def _device_count():
    return torch.cuda.device_count()


def average_several_run(run_func, args, n_times=4, n_paral=2, **run_params):
    """Get average result after several running.

    Args:
        run_func [func]: running function like 'Classify.train_test'
        args [dict]: all model arguments
        n_times [int]: run several times for average
        n_paral [int]: number of parallel processes
        run_params [dict]: parameters for runing function

    Returns:
        max_socres [dict]: dict of the maximum scores after n_times running
    """
    dnnnlp.verbose.config(0)

    assert not n_times % n_paral, ValueError("'n_times' should be an integral multiple of 'n_paral'.")

    pool = mp.Pool(processes=1)
    check = pool.apply_async(_device_count)
    device_count = check.get()

    scores, processes = [], []
    pool = mp.Pool(processes=n_paral)

    if args.n_gpu > 0:
        assert n_paral * args.n_gpu <= device_count, "Not enough GPU devices."

    for t in range(n_times):
        if args.n_gpu > 0:
            run_params['device_id'] = (t % n_paral) * args.n_gpu
        else:
            run_params['device_id'] = -1

        processes.append(pool.apply_async(
            run_func,
            kwds=run_params.copy()
        ))

        if (t + 1) % n_paral == 0:
            for i, p in enumerate(processes):
                result = p.get()
                scores.append(result)

                print()
                ptable = utils.display_prfacc(args.eval_metric, verbose=0)
                ptable.row(dict(result, **{"iter": t + 2 - len(processes) + i}))
            processes.clear()

    avg_scores = utils.average_prfacc(*scores)
    print()
    ptable = utils.display_prfacc(args.eval_metric, verbose=0)
    ptable.row(dict(avg_scores, **{"iter": 'AVG'}))

    dnnnlp.verbose.config(2)
    return avg_scores


def _pack_run(exec_class, run_func, **run_params):
    time.sleep(1)
    params_set = run_params.pop('params_set')
    exec_class.attributes_from_dict(params_set)
    results = run_func(**run_params)
    return results


def grid_search(exec_class, run_func, args, params_search, n_paral=2, **run_params):
    """Do parameters' grid search.

    Args:
        exec_func [exec]: an exec class like 'Classify'
        run_func [func]: running function like 'Classify.train_test'
        args [dict]: all model arguments
        params_search [dict]: grid search parameters
        n_paral [int]: number of parallel processes
        run_params [dict]: parameters for run function

    Returns:
        max_socres [dict]: dict of the maximum scores after n_times running
    """
    from sklearn.model_selection import ParameterGrid

    dnnnlp.verbose.config(0)

    pool = mp.Pool(processes=1)
    check = pool.apply_async(_device_count)
    device_count = check.get()

    scores, processes = [], []
    pool = mp.Pool(processes=n_paral)

    if args.n_gpu > 0:
        assert n_paral * args.n_gpu <= device_count, "Not enough GPU devices."

    def get_result(t):
        for i, p in enumerate(processes):
            result = p.get()
            scores.append(result)
            print()
            print(params_search[t + 1 -len(processes) + i])
            ptable = utils.display_prfacc(args.eval_metric, verbose=0)
            ptable.row(dict(result, **{"iter": t + 2 - len(processes) + i}))

    params_search = list(ParameterGrid(params_search))
    for t, params in enumerate(params_search):
        run_params['params_set'] = params
        if args.n_gpu > 0:
            run_params['device_id'] = (t % n_paral) * args.n_gpu
        else:
            run_params['device_id'] = -1

        processes.append(pool.apply_async(
            _pack_run,
            args=(exec_class, run_func),
            kwds=run_params.copy()
        ))

        if (t + 1) % n_paral == 0:
            get_result(t)
            processes.clear()

    get_result(t)

    max_scores = utils.maximum_prfacc(*scores, eval_metric=args.eval_metric)
    print()
    ptable = utils.display_prfacc(args.eval_metric, verbose=0)
    ptable.row(dict(max_scores, **{"iter": 'MAX'}))

    dnnnlp.verbose.config(2)
    return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Some common layers for deep neural network.
Last update: KzXuan, 2019.08.21
"""
import math
import torch
import numpy as np
import torch.nn as nn
import torch.utils.data as Data
import torch.nn.functional as F


class EmbeddingLayer(nn.Module):
    def __init__(self, emb_matrix, emb_type='const'):
        """Initilize embedding layer.

        Args:
            emb_matrix [tensor/np.array]: embedding matrix
            emb_type [str]: use None/'const'/'variable' to set the layer
        """
        super(EmbeddingLayer, self).__init__()

        # embedding will not be changed
        if emb_type == 'const':
            self.emb_mat = nn.Embedding.from_pretrained(torch.FloatTensor(emb_matrix))
        # embedding will be changed
        elif emb_type == 'variable':
            self.emb_mat = nn.Embedding.from_pretrained(torch.FloatTensor(emb_matrix), freeze=False)
        # no embedding
        elif emb_type is None:
            self.emb_mat = None
        else:
            raise "Value error of 'emb_type', wants None/'const'/'variable', gets '{}'.".format(emb_type)
        self.emb_type = emb_type

    def forward(self, inputs):
        """Forward propagation.

        Args:
            inputs [tensor]: input tensor (batch_size * max_seq_len)

        Returns:
            After-embedding-inputs or original-inputs (batch_size * max_seq_len * emb_dim)
        """
        if self.emb_type is not None:
            return self.emb_mat(inputs.long())
        else:
            return inputs


class SoftmaxLayer(nn.Module):
    def __init__(self, input_size, output_size):
        """Initilize softmax layer / full connected layer.

        Args:
            input_size [int]: number of input nodes
            output_size [int]: number of output nodes
        """
        super(SoftmaxLayer, self).__init__()

        self.input_size = input_size
        self.connector = nn.Sequential(
            nn.Linear(input_size, output_size),
            nn.LogSoftmax(-1)
        )

    def forward(self, inputs):
        """Forward propagation.

        Args:
            inputs [tensor]: inputs tensor (batch_size * ... * input_size)

        Returns:
            outputs [tensor]: output tensor (batch_size * ... * output_size)
        """
        assert inputs.size(-1) == self.input_size, "Dimension error of 'inputs'."
        outputs = self.connector(inputs)
        return outputs


class SoftAttentionLayer(nn.Module):
    def __init__(self, input_size):
        """Initilize soft attention layer.

        Args:
            input_size [int]: embedding dim or the last dim of the input
        """
        super(SoftAttentionLayer, self).__init__()

        self.input_size = input_size
        self.attention = nn.Sequential(
            nn.Linear(input_size, input_size),
            nn.Tanh(),
            nn.Linear(input_size, 1, bias=False),
        )

    def forward(self, inputs, mask=None):
        """Forward propagation.

        Args:
            inputs [tensor]: input tensor (batch_size * max_seq_len * input_size)
            mask [tensor]: mask matrix (batch_size * max_seq_len)

        Returns:
            outputs [tensor]: output tensor (batch_size * input_size)
        """
        assert inputs.dim() == 3, "Dimension error of 'inputs'."
        assert inputs.size(-1) == self.input_size, "Dimension error of 'inputs'."
        assert mask is None or mask.dim() == 2, "Dimension error of 'mask'."

        now_batch_size, max_seq_len, _ = inputs.size()
        alpha = self.attention(inputs).reshape(now_batch_size, 1, max_seq_len)
        exp = alpha.exp()

        if mask is not None:
            exp = exp * mask.unsqueeze(1).float()

        sum_exp = exp.sum(-1, keepdim=True) + 1e-9
        softmax_exp = exp / sum_exp.expand_as(exp).reshape(now_batch_size, 1, max_seq_len)
        outputs = torch.bmm(softmax_exp, inputs).squeeze()
        return outputs


class CNNLayer(nn.Module):
    def __init__(self, input_size, in_channels, out_channels,
                 kernel_width, act_fun=nn.ReLU, drop_prob=0.1):
        """Initilize CNN layer.

        Args:
            input_size [int]: embedding dim or the last dim of the input
            in_channels [int]: number of channels for inputs
            out_channels [int]: number of channels for outputs
            kernel_width [int]: the width on sequence for the first dim of kernel
            act_fun [torch.nn.modules.activation]: activation function
            drop_prob [float]: drop out ratio
        """
        super(CNNLayer, self).__init__()

        self.input_size = input_size
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_width = kernel_width

        self.conv = nn.Conv2d(in_channels, out_channels, (kernel_width, input_size))
        self.drop_out = nn.Dropout(drop_prob)

        assert callable(act_fun), TypeError("Type error of 'act_fun', use functions like nn.ReLU/nn.Tanh.")
        self.act_fun = act_fun()

    def forward(self, inputs, mask=None, out_type='max'):
        """Forward propagation.

        Args:
            inputs [tensor]: input tensor (batch_size * in_channels * max_seq_len * input_size)
                             or (batch_size * max_seq_len * input_size)
            mask [tensor]: mask matrix (batch_size * max_seq_len)
            out_type [str]: use 'max'/'mean'/'all' to choose

        Returns:
            outputs [tensor]: output tensor (batch_size * out_channels) or (batch_size * left_len * n_hidden)
        """
        # auto extend 3d inputs
        if inputs.dim() == 3:
            inputs = inputs.unsqueeze(1).repeat(1, self.in_channels, 1, 1)
        assert inputs.dim() == 4 and inputs.size(1) == self.in_channels, "Dimension error of 'inputs'."
        assert inputs.size(-1) == self.input_size, "Dimension error of 'inputs'."

        now_batch_size, _, max_seq_len, _ = inputs.size()
        assert max_seq_len >= self.kernel_width, "Dimension error of 'inputs'."
        assert out_type in ['max', 'mean', 'all'], ValueError(
            "Value error of 'out_type', only accepts 'max'/'mean'/'all'."
        )

        # calculate the seq_len after convolution
        left_len = max_seq_len - self.kernel_width + 1

        # auto generate full-one mask
        if mask is None:
            mask = torch.ones((now_batch_size, left_len), device=inputs.device)
        assert mask.dim() == 2, "Dimension error of 'mask'."
        mask = mask[:, -left_len:].unsqueeze(1)

        outputs = self.conv(inputs)
        outputs = self.drop_out(outputs)
        outputs = outputs.reshape(-1, self.out_channels, left_len)

        outputs = self.act_fun(outputs)  # batch_size * out_channels * left_len

        # all modes need to consider mask
        if out_type == 'max':
            outputs = outputs.masked_fill(~mask.bool(), -1e10)
            outputs = F.max_pool1d(outputs, left_len).reshape(-1, self.out_channels)
            isinf = outputs.eq(-1e10)
            outputs = outputs.masked_fill(isinf, 0)
        elif out_type == 'mean':
            outputs = outputs.masked_fill(~mask.bool(), 0)
            lens = torch.sum(mask, dim=-1)
            outputs = torch.sum(outputs, dim=-1) / (lens.float() + 1e-9)
        elif out_type == 'all':
            outputs = outputs.masked_fill(~mask.bool(), 0)
            outputs = outputs.transpose(1, 2)  # batch_size * left_len * out_channels
        return outputs


class RNNLayer(nn.Module):
    def __init__(self, input_size, n_hidden, n_layer=1, bi_direction=True,
                 rnn_type='LSTM', drop_prob=0.1):
        """Initilize RNN layer.

        Args:
            input_size [int]: embedding dim or the last dim of the input
            n_hidden [int]: number of hidden layer nodes
            n_layer [int]: number of hidden layers
            bi_direction [bool]: use bi-directional model or not
            rnn_type [str]: choose rnn type with 'tanh'/'LSTM'/'GRU'
            drop_prob [float]: drop out ratio
        """
        super(RNNLayer, self).__init__()

        models = {'tanh': nn.RNN, 'LSTM': nn.LSTM, 'GRU': nn.GRU}
        assert rnn_type in models.keys(), ValueError("Value error of 'mode', only accepts 'tanh'/'LSTM'/'GRU'.")

        self.input_size = input_size
        self.n_hidden = n_hidden
        self.n_layer = n_layer
        self.bi_direction_num = 2 if bi_direction else 1
        self.rnn_type = rnn_type
        self.rnn = models[rnn_type](
            input_size=input_size,
            hidden_size=n_hidden,
            num_layers=n_layer,
            bias=True,
            batch_first=True,
            dropout=drop_prob if n_layer > 1 else 0,
            bidirectional=bi_direction
        )

    def forward(self, inputs, mask=None, out_type='last'):
        """Forward propagation.

        Args:
            inputs [tensor]: input tensor (batch_size * max_seq_len * input_size)
            mask [tensor]: mask matrix (batch_size * max_seq_len)
            out_type [str]: use 'all'/'last' to choose

        Returns:
            h_last [tensor]: the last time step output tensor (batch_size * (bi_direction * n_hidden))
            outputs [tensor]: the last layer output tensor (batch_size * max_seq_len * (bi_direction * n_hidden))
        """
        assert inputs.dim() == 3, "Dimension error of 'inputs'."
        assert inputs.size(-1) == self.input_size, "Dimension error of 'inputs'."
        assert mask is None or mask.dim() == 2, "Dimension error of 'mask'."
        assert out_type in ['last', 'all'], ValueError("Value error of 'out_type', only accepts 'last'/'all'.")

        # convert mask to sequence length
        if mask is not None:
            seq_len = mask.sum(dim=-1).int()
        else:
            seq_len = torch.ones((inputs.size(0),))
        # remove full-zero rows
        nonzero_index = seq_len.nonzero().reshape(-1)
        zero_index = (seq_len == 0).nonzero().reshape(-1)
        # get rebuild index
        _, re_index = torch.cat((nonzero_index, zero_index)).sort(descending=False)

        inputs = inputs.index_select(0, nonzero_index)
        seq_len = seq_len.index_select(0, nonzero_index)
        now_batch_size, max_seq_len, _ = inputs.size()

        self.rnn.flatten_parameters()
        inputs = nn.utils.rnn.pack_padded_sequence(inputs, seq_len, batch_first=True, enforce_sorted=False)
        if self.rnn_type == 'tanh' or self.rnn_type == 'GRU':
            outputs, h_last = self.rnn(inputs)
        elif self.rnn_type == 'LSTM':
            outputs, (h_last, _) = self.rnn(inputs)

        if out_type == 'last':
            h_last = h_last.reshape(self.n_layer, self.bi_direction_num, now_batch_size, self.n_hidden)
            h_last = h_last[-1].transpose(0, 1).reshape(now_batch_size, self.bi_direction_num * self.n_hidden)
            # pad full-zero rows back and rebuild
            h_last = F.pad(h_last, (0, 0, 0, zero_index.size(0)))
            h_last = h_last.index_select(0, re_index)
            return h_last  # batch_size * (bi_direction * n_hidden)
        elif out_type == 'all':
            outputs, _ = nn.utils.rnn.pad_packed_sequence(outputs, batch_first=True, total_length=max_seq_len)
            # pad full-zero rows back and rebuild
            outputs = F.pad(outputs, (0, 0, 0, 0, 0, zero_index.size(0)))
            outputs = outputs.index_select(0, re_index)
            return outputs  # batch_size * max_seq_len * (bi_direction * n_hidden)


class CRFLayer(nn.Module):
    def __init__(self):
        super(CRFLayer, self).__init__()


class MultiheadAttentionLayer(nn.Module):
    def __init__(self, input_size, n_head=8, drop_prob=0.1):
        """Initilize multi-head attention layer.

        Args:
            input_size [int]: embedding dim or the last dim of the input
            n_head [int]: number of attention heads
            drop_prob [float]: drop out ratio
        """
        super(MultiheadAttentionLayer, self).__init__()

        self.input_size = input_size
        self.attention = nn.MultiheadAttention(input_size, n_head, drop_prob)

    def forward(self, query, key=None, value=None, query_mask=None, key_mask=None):
        """Forward propagation.

        Args:
            query [tensor]: query tensor (batch_size * max_seq_len_query * input_size)
            key [tensor]: key tensor (batch_size * max_seq_len_key * input_size)
            value [tensor]: value tensor (batch_size * max_seq_len_key * input_size)
            query_mask [tensor]: query mask matrix (batch_size * max_seq_len_query)
            key_mask [tensor]: key mask matrix (batch_size * max_seq_len_key)

        Returns:
            outputs [tensor]: output tensor (batch_size * max_seq_len_query * input_size)
        """
        assert query.dim() == 3, "Dimension error of 'query'."
        assert query.size(-1) == self.input_size, "Dimension error of 'query'."
        # set key = query
        if key is None:
            key, key_mask = query, query_mask
        assert key.dim() == 3, "Dimension error of 'key'."
        assert key.size(-1) == self.input_size, "Dimension error of 'key'."
        # set value = key
        value = key if value is None else value
        assert value.dim() == 3, "Dimension error of 'value'."
        assert value.size(-1) == self.input_size, "Dimension error of 'value'."

        assert query.size(0) == key.size(0) == value.size(0), "Dimension match error."
        assert key.size(1) == value.size(1), "Dimension match error of 'key' and 'value'."
        assert query.size(2) == key.size(2) == value.size(2), "Dimension match error."

        if query_mask is not None:
            assert query_mask.dim() == 2, "Dimension error of 'query_mask'."
            assert query_mask.shape == query.shape[:2], "Dimension match error of 'query' and 'query_mask'."

        # auto generate full-one mask
        if key_mask is None:
            key_mask = torch.ones(key.shape[:2], device=query.device)
        assert key_mask.dim() == 2, "Dimension error of 'key_mask'."
        assert key_mask.shape == key.shape[:2], "Dimension match error of 'key' and 'key_mask'."

        # transpose dimension batch_size and max_seq_len
        query, key, value = query.transpose(0, 1), key.transpose(0, 1), value.transpose(0, 1)
        outputs, _ = self.attention(query, key, value, key_padding_mask=~key_mask.bool())
        # transpose back
        outputs = outputs.transpose(0, 1)
        outputs = outputs.masked_fill(torch.isnan(outputs), 0)
        if query_mask is not None:
            query_mask = query_mask.unsqueeze(2)
            outputs = outputs.masked_fill(~query_mask.bool(), 0)
        return outputs


class TransformerLayer(nn.Module):
    def __init__(self, input_size, n_head=8, feed_dim=None, drop_prob=0.1):
        """Initilize transformer layer.

        Args:
            input_size [int]: embedding dim or the last dim of the input
            n_head [int]: number of attention heads
            feed_dim [int]: hidden matrix dimension
            drop_prob [float]: drop out ratio
        """
        super(TransformerLayer, self).__init__()

        self.input_size = input_size
        self.feed_dim = 4 * self.input_size if feed_dim is None else feed_dim

        self.attention = MultiheadAttentionLayer(input_size, n_head, drop_prob)
        self.drop_out_1 = nn.Dropout(drop_prob)
        self.norm_1 = nn.LayerNorm(input_size)
        self.linear_1 = nn.Linear(input_size, self.feed_dim)
        self.drop_out_2 = nn.Dropout(drop_prob)
        self.linear_2 = nn.Linear(self.feed_dim, input_size)
        self.drop_out_3 = nn.Dropout(drop_prob)
        self.norm_2 = nn.LayerNorm(input_size)

    def forward(self, query, key=None, value=None, query_mask=None,
                key_mask=None, out_type='first'):
        """Forward propagation.

        Args:
            query [tensor]: query tensor (batch_size * max_seq_len_query * input_size)
            key [tensor]: key tensor (batch_size * max_seq_len_key * input_size)
            value [tensor]: value tensor (batch_size * max_seq_len_key * input_size)
            query_mask [tensor]: query mask matrix (batch_size * max_seq_len_query)
            key_mask [tensor]: key mask matrix (batch_size * max_seq_len_key)
            out_type [str]: use 'first'/'all' to choose

        Returns:
            outputs [tensor]: output tensor (batch_size * input_size)
                              or (batch_size * max_seq_len_query * input_size)
        """
        assert query.dim() == 3, "Dimension error of 'query'."
        assert query.size(-1) == self.input_size, "Dimension error of 'query'."
        assert out_type in ['first', 'all'], ValueError(
            "Value error of 'out_type', only accepts 'first'/'all'."
        )

        outputs = self.attention(query, key, value, query_mask, key_mask)
        # residual connection
        outputs = query + self.drop_out_1(outputs)
        outputs = self.norm_1(outputs)

        temp = self.linear_2(self.drop_out_2(F.relu(self.linear_1(outputs))))
        # residual connection
        outputs = outputs + self.drop_out_3(temp)
        outputs = self.norm_2(outputs)

        if query_mask is not None:
            query_mask = query_mask.unsqueeze(2)
            outputs = outputs.masked_fill(~query_mask.bool(), 0)

        if out_type == 'first':
            outputs = outputs[:, 0, :]
        return outputs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import dnnnlp
from setuptools import setup, find_packages

with open("readme.md", "r", encoding='utf-8') as fobj:
    long_description = fobj.read()

setup(
    name="dnnnlp",
    version="1.0.3",
    author="KzXuan",
    author_email="kaizhouxuan@gmail.com",
    description="Deep Neural Networks for Natural Language Processing classification or sequential task written by PyTorch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    url="https://github.com/NUSTM/pytorch-dnnnlp",
    packages=find_packages(),
    install_requires=[
        "torch >= 1.2.0",
        ],
    classifiers=[
        "License :: Free for non-commercial use",
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Some utilities for deep neural network.
Last update: KzXuan, 2019.08.22
"""
import sys
import dnnnlp
import numpy as np
from sklearn.metrics import classification_report
np.seterr(divide='ignore', invalid='ignore')

def sysprint(_str):
    """Print without '\n'.

    Args:
        _str [str]: string to output
    """
    sys.stdout.write(_str)
    sys.stdout.flush()


def mod_fold(length, fold=10):
    """Use mod index to fold.

    Args:
        length [int]: length of data
        fold [int]: fold in need (default=10)

    Returns:
        indexs [list]: [(fold_num, train_index, test_index),]
    """
    indexs = []
    _all = np.arange(length, dtype=int)
    for f in range(fold):
        test = np.arange(f, length, fold, dtype=int)
        train = np.setdiff1d(_all, test)
        indexs.append((f + 1, train, test))
    return indexs


def order_fold(length, fold=10):
    """Use order index to fold.

    Args:
        length [int]: length of data
        fold [int]: fold in need (default=10)

    Returns:
        indexs [list]: [(fold_num, train_index, test_index),]
    """
    indexs = []
    _all = np.arange(length, dtype=int)
    gap, left = length // fold, length % fold
    begin = 0
    for f in range(fold):
        step = gap + 1 if f < left else gap
        test = np.arange(begin, begin + step, dtype=int)
        train = np.setdiff1d(_all, test)
        begin += step
        indexs.append((f + 1, train, test))
    return indexs


def one_hot(arr, n_class=0):
    """Change labels to one-hot expression.

    Args:
        arr [np.array]: numpy array
        n_class [int]: number of class

    Returns:
        oh [np.array]: numpy array with one-hot expression
    """
    if arr is None:
        return None
    if isinstance(arr, list) or isinstance(arr, np.ndarray):
        arr = np.array(arr)
        ishape = arr.shape
        arr = arr.flatten()

        n_class = arr.max() + 1 if n_class == 0 else n_class
        assert n_class >= arr.max() + 1, ValueError("Value of 'n_class' is too small.")

        oh = np.zeros((arr.size, n_class), dtype=int)
        oh[np.arange(arr.size), arr] = 1
        oh = np.reshape(oh, (*ishape, -1))

    return oh


def len_to_mask(seq_len, max_seq_len=None):
    """Convert seq_len to mask matrix.

    Args:
        seq_len [tensor]: sequence length vector (batch_size,)
        max_seq_len [int]: max sequence length

    Returns:
        mask [tensor]: mask matrix (batch_size * max_seq_len)
    """
    if isinstance(seq_len, np.ndarray):
        if max_seq_len is None:
            max_seq_len = seq_len.max()
        query = np.arange(0, max_seq_len)
        mask = (query < seq_len.reshape(-1, 1)).astype(int)
    else:
        import torch
        if max_seq_len is None:
            max_seq_len = seq_len.max()
        query = torch.arange(0, max_seq_len, device=seq_len.device).float()
        mask = torch.lt(query, seq_len.unsqueeze(1)).int()
    return mask


def mask_to_len(mask):
    """Convert mask matrix to seq_len.

    Args:
        mask [tensor]: mask matrix (batch_size * max_seq_len)

    Returns:
        seq_len [tensor]: sequence length vector (batch_size,)
    """
    if isinstance(mask, np.ndarray):
        seq_len = np.sum(mask, axis=1).astype(int)
    else:
        import torch
        seq_len = mask.sum(dim=1).int()
    return seq_len


def _form_digits(evals, ndigits):
    """Form digits in dict.

    Args:
        evals [dict]: dict of all the evaluation metrics
        ndigits [int]: decimal number of float
    """
    for key, value in evals.items():
        if isinstance(value, float):
            evals[key] = round(float(value), ndigits)
        if isinstance(value, dict):
            _form_digits(value, ndigits)


def prfacc(y_true, y_pred, mask=None, one_hot=False, ndigits=4, tabular=False):
    """Evaluation of true label and prediction.

    Args:
        y_true [np.array/list]: true label size of (n_sample, *) / (n_sample, *, n_class)
        y_pred [np.array/list]: predict label size of (n_sample, *) / (n_sample, *, n_class)
        one_hot [bool]: True for (n_sample, *, n_class) input
        ndigits [int]: decimal number of float
        tabular [bool]: return a table

    Returns:
        evals [dict/str]: dict of all the evaluation metrics or report table
    """
    y_true = np.array(y_true) if isinstance(y_true, list) else y_true
    y_pred = np.array(y_pred) if isinstance(y_pred, list) else y_pred
    mask = np.array(mask) if isinstance(mask, list) else mask

    if not isinstance(y_true, np.ndarray):
        try:
            import torch
            y_true = y_true.cpu().data.numpy() if isinstance(y_true, torch.Tensor) else y_true
            y_pred = y_pred.cpu().data.numpy() if isinstance(y_pred, torch.Tensor) else y_pred
            mask = mask.cpu().data.numpy() if isinstance(mask, torch.Tensor) else mask
        except:
            TypeError("Type error of the input matrices.")

    assert y_true.ndim == y_pred.ndim, "Dimension match error."
    if one_hot:
        y_true = np.argmax(y_true, axis=-1)
        y_pred = np.argmax(y_pred, axis=-1)

    if mask is not None:
        assert y_true.shape == y_pred.shape == mask.shape, "Dimension error."
        mask_ind = np.where(mask == 1)
        y_true, y_pred = y_true[mask_ind], y_pred[mask_ind]
    else:
        y_true, y_pred = y_true.flatten(), y_pred.flatten()

    n_class = max(np.max(y_true), np.max(y_pred)) + 1
    names = ['class{}'.format(i) for i in range(n_class)]

    evals = classification_report(
        y_true, y_pred, digits=ndigits, target_names=names, output_dict=not tabular
    )

    if not tabular:
        evals['macro'] = evals.pop("macro avg")
        evals['weighted'] = evals.pop("weighted avg")
        p, r = evals['macro']['precision'], evals['macro']['recall']
        evals['macro']['f1-score'] = 2 * p * r / (p + r)
        _form_digits(evals, ndigits)

    return evals


def average_prfacc(*evals, ndigits=4):
    """Average for multiple evaluations.

    Args:
        evals [tuple]: several evals without limitation
        ndigits [int]: decimal number of float

    Returns:
        avg [dict]: dict of the average values of all the evaluation metrics
    """
    avg = {}.fromkeys(evals[0].keys())
    for key in avg:
        values = [e[key] for e in evals]
        if isinstance(values[0], dict):
            avg[key] = average_prfacc(*values)
        else:
            avg[key] = round(sum(values) / len(values), ndigits)

    return avg


def maximum_prfacc(*evals, eval_metric='accuracy'):
    """Get maximum for multiple evaluations.

    Args:
        evals [tuple]: several evals without limitation
        eval_metric [str]: evaluation standard for comparsion

    Returns:
        max_eval [dict]: one eval with the maximum score
    """
    assert eval_metric in evals[0].keys(), ValueError("Value error of 'eval_metric'.")

    if eval_metric == 'accuracy':
        values = [e[eval_metric] for e in evals]
    else:
        values = [e[eval_metric]['f1-score'] for e in evals]
    index = values.index(max(values))
    max_eval = evals[index]

    return max_eval


class display_prfacc(object):
    """Display evaluations line by line.
    """
    def __init__(self, *eval_metrics, sep='|', verbose=2):
        """Initilize and print head.

        Args:
            eval_metrics [str]: several wanted evaluation metrics
            sep [str]: separate mark like ' '/'|'/'*'
            verbose [int]: verbose level
        """
        self.verbose = verbose
        if not dnnnlp.verbose.check(self.verbose):
            return
        eval_metrics = list(eval_metrics)
        for i, em in enumerate(eval_metrics):
            if em[:5] not in ['accur', 'macro', 'micro', 'class']:
                raise ValueError("Value error of the 'eval_metric'.")
            if em in ['accuracy', 'micro']:
                eval_metrics[i] = 'macro'
        self.eval_metrics = list(set(eval_metrics))
        self.eval_metrics.sort(key=eval_metrics.index)

        self.col = ["iter", "loss", "acc", *self.eval_metrics]
        self.width = [4, 6, 6] + [22] * len(self.eval_metrics)
        self.sep = ' ' + sep + ' '

        for i in range(len(self.col)):
            sysprint(self.sep)
            sysprint("{:^{}}".format(self.col[i], self.width[i]))
        print(self.sep)

    def line(self):
        """Print a line.
        """
        if not dnnnlp.verbose.check(self.verbose):
            return
        for i in range(len(self.col)):
            sysprint(self.sep)
            sysprint("-" * self.width[i])
        print(self.sep)

    def row(self, evals):
        """Process and print a row.
        Atgs:
            evals [dict]: dict of all the evaluation metrics
        """
        if not dnnnlp.verbose.check(self.verbose):
            return
        sysprint(self.sep)
        sysprint("{:^4}".format(evals.get('iter', '-')))
        sysprint(self.sep)
        if 'loss' in evals:
            sysprint("{:^.4f}".format(evals['loss'])[:6])
        else:
            sysprint("{:^6}".format('-'))
        sysprint(self.sep)
        sysprint("{:^.4f}".format(evals['accuracy'])[:6])
        for em in self.eval_metrics:
            sysprint(self.sep)
            value = evals.get(em, '-')
            if value != '-':
                value = "{:.4f}".format(value['precision'])[:6] + "  " +\
                        "{:.4f}".format(value['recall'])[:6] + "  " +\
                        "{:.4f}".format(value['f1-score'])[:6]
            sysprint("{:^22}".format(value))
        print(self.sep)
